/// Copyright (c) 2005 Gabriele Bartolini, Prato, Italia
///
/// This program is free software; you can redistribute it and/or modify
/// it under the terms of the GNU General Public License as published by
/// the Free Software Foundation; either version 2 of the License, or
/// (at your option) any later version.
///
/// This program is distributed in the hope that it will be useful,
/// but WITHOUT ANY WARRANTY; without even the implied warranty of
/// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
/// GNU General Public License for more details.
///
/// You should have received a copy of the GNU General Public License
/// along with this program; if not, write to the Free Software
/// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
///
///////////////////////////////////////////////////////////////
/// \version $Id: Memento.h,v 1.3 2005/03/05 14:11:30 angusgb Exp $
///////////////////////////////////////////////////////////////

#ifndef __Memento_H
#define __Memento_H

/// \class Memento Memento.h Memento.h
/// \brief Memento pattern (GoF)
///
/// This template class implements the Memento pattern
/// as described in the GoF book. Will be used to perform
/// Undo/Redo operations or back/forward operations

template <class T>
class Memento{
public:
    inline Memento(T& s);

    /// Copy constructor
    /// \param rhs Object to be copied (right hand side)
    inline Memento(const Memento& rhs);

    ~Memento();

    inline void setState(T& s);
    inline T* getState();

    /// Assignment operator
    /// \param rhs Object to be assigned (right hand side)
    Memento& operator=(const Memento& rhs);

    /// Compare a Memento object with another one
    /// \param rhs Object to be compared (right hand side)
    inline int compare (const Memento& rhs) const;

    /// Equality operator
    /// \param rhs Object to be compared (right hand side)
    bool operator==(const Memento& rhs) const;

    /// Less than operator
    /// \param rhs Object to be compared (right hand side)
    bool operator<(const Memento& rhs) const;

    /// Diversity than operator
    /// \param rhs Object to be compared (right hand side)
    bool operator!=(const Memento& rhs) const {return !operator==(rhs);}

    /// Greater than operator
    /// \param rhs Object to be compared (right hand side)
    bool operator>(const Memento& rhs) const { return !(rhs < *this);}

protected:
    Memento();

    T* _state; ///< State of the originator object
};

///////////////////////////////////////////////////////////////
// Implementation
///////////////////////////////////////////////////////////////
// Main constructor
template <class T>
Memento<T>::Memento(T& s)
: _state(&s){
}

// Default constructor
template <class T>
Memento<T>::Memento()
    : _state(0){
}

// Copy constructor
template <class T>
Memento<T>::Memento(const Memento& rhs)
    : _state (rhs._state){
}

// Destructor
template <class T>
Memento<T>::~Memento(){
}

// Assignment operator
template <class T>
Memento<T>& Memento<T>::operator=(const Memento& rhs){
    if (this == &rhs)
        return *this;

    // Code for attributes copy
    _state = rhs._state;

    return *this; // returns itself
}

// Compare the Memento object
template <class T>
int Memento<T>::compare(const Memento& rhs) const{
    return (*_state).compare(*(rhs._state));
}

// Equality operator
template <class T>
bool Memento<T>::operator==(const Memento& rhs) const{
    return !compare(rhs);
}

// Less than operator
template <class T>
bool Memento<T>::operator<(const Memento& rhs) const{
    return (compare(rhs) < 0);
}

// Set the state
template <class T>
void Memento<T>::setState(T& s){
    _state = &s;
}

// Get the state
template <class T>
T* Memento<T>::getState(){
    return _state;
}

#endif