///////////////////////////////////////////////////////////////
///
/// Copyright (c) 2005 Gabriele Bartolini, Prato, Italia
///
/// This program is free software; you can redistribute it and/or modify
/// it under the terms of the GNU General Public License as published by
/// the Free Software Foundation; either version 2 of the License, or
/// (at your option) any later version.
///
/// This program is distributed in the hope that it will be useful,
/// but WITHOUT ANY WARRANTY; without even the implied warranty of
/// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
/// GNU General Public License for more details.
///
/// You should have received a copy of the GNU General Public License
/// along with this program; if not, write to the Free Software
/// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
///
///////////////////////////////////////////////////////////////
/// \version $Id: Originator.h,v 1.3 2005/03/05 14:11:30 angusgb Exp $
///////////////////////////////////////////////////////////////

#ifndef __Originator_H
#define __Originator_H

#ifndef __Memento_H
#include "Memento.h"
#endif

/// \class Originator Originator.h Originator.h
/// \brief Originator, from the Memento pattern (GoF)
///
/// This template class implements the Originator pattern
/// as described in the GoF book. Will be used to perform
/// Undo/Redo operations or back/forward operations

template <class T>
class Originator{

// Public interface
public:

    Originator(const T& s);

    /// Copy constructor
    /// \param rhs Object to be copied (right hand side)
    Originator(const Originator& rhs);

    ~Originator();

    inline void setMemento(Memento<T>& m);
    inline Memento<T>* createMemento();

    /// Assignment operator
    /// \param rhs Object to be assigned (right hand side)
    Originator& operator=(const Originator& rhs);

    /// Compare a Originator object with another one
    /// \param rhs Object to be compared (right hand side)
    inline int compare (const Originator& rhs) const;

    /// Equality operator
    /// \param rhs Object to be compared (right hand side)
    bool operator==(const Originator& rhs) const;

    /// Less than operator
    /// \param rhs Object to be compared (right hand side)
    bool operator<(const Originator& rhs) const;

    /// Diversity than operator
    /// \param rhs Object to be compared (right hand side)
    bool operator!=(const Originator& rhs) const {return !operator==(rhs);}

    /// Greater than operator
    /// \param rhs Object to be compared (right hand side)
    bool operator>(const Originator& rhs) const { return !(rhs < *this);}

// Protected attributes of the class
protected:
    Originator();

    T* _state; ///< Current state

};

///////////////////////////////////////////////////////////////
// Implementation
///////////////////////////////////////////////////////////////
// Main constructor
template <class T>
Originator<T>::Originator(const T& s)
    : _state(new T(s)){
}

// Default constructor
template <class T>
Originator<T>::Originator()
    : _state(0){
}

// Copy constructor
template <class T>
Originator<T>::Originator(const Originator& rhs)
    : _state (new T(* (rhs._state))){
}

// Destructor
template <class T>
Originator<T>::~Originator(){
}

// Assignment operator
template <class T>
Originator<T>& Originator<T>::operator=(const Originator& rhs){
    if (this == &rhs)
        return *this;

    // Code for attributes copy
    _state = new T (* (rhs._state));

    return *this; // returns itself
}

// Compare the Originator object
template <class T>
int Originator<T>::compare(const Originator& rhs) const{
    return (*_state).compare(*(rhs._state));
}

// Equality operator
template <class T>
bool Originator<T>::operator==(const Originator& rhs) const{
    return !compare(rhs);
}

// Less than operator
template <class T>
bool Originator<T>::operator<(const Originator& rhs) const{
    return (compare(rhs) < 0);
}

// Set the state
template <class T>
void Originator<T>::setMemento(Memento<T>& m){
    _state = m.getState();
}

// Get the state
template <class T>
Memento<T>* Originator<T>::createMemento(){
    return new Memento<T>(*_state);
}

#endif