#include "osgHelperFunctions.h"
#include "TrainerAnimation.h"

#include "Trainer/Athlete.h"
#include "Trainer/Trainer.h"
#include "Trainer/DataVisitor.h"
#include "Trainer/XMLVisitor.h"

#include <osgART/Foundation>
#include <osgART/VideoLayer>
#include <osgART/PluginManager>
#include <osgART/VideoGeode>
#include <osgART/Utils>
#include <osgART/GeometryUtils>
#include <osgART/MarkerCallback>
#include <osgART/TransformFilterCallback>
#include <osgViewer/Viewer>
#include <osgViewer/ViewerEventHandlers>

#include <cstdlib>
#include <osg/Geode>
#include <osgViewer/Viewer>
#include <osgText/Font>
#include <osgText/Text>

//#define NDEBUG 1

#ifndef _PI
#define _PI           3.14159265358979323846
#endif

using namespace std;
using namespace osg;

osgViewer::Viewer viewer;
ref_ptr<MatrixTransform> transformMatrix = new MatrixTransform();
ref_ptr<osgART::Marker> marker;
osgText::Text *hud;
TrainerAnimation trainerAnimation;
ref_ptr<osgART::Video> video;

static Group* createImageBackground(Image* video) {
	osgART::VideoLayer* _layer = new osgART::VideoLayer();
	_layer->setSize(*video);
	osgART::VideoGeode* _geode = new osgART::VideoGeode(osgART::VideoGeode::USE_TEXTURE_2D, video);
	addTexturedQuad(*_geode,video->s(),video->t());

	hud=osgHelperFunctions::createLabel("Test", 12);
	hud->setDataVariance(osg::Object::DYNAMIC);
	_geode->addDrawable(hud);
	_layer->addChild(_geode);
	return _layer;
}

void setUpGraphics(){
	// create a root node
	ref_ptr<Group> root = new Group;

	viewer.setSceneData(root.get());

	viewer.addEventHandler(new osgViewer::StatsHandler);
	viewer.addEventHandler(new osgViewer::WindowSizeHandler);
	viewer.addEventHandler(new osgViewer::ThreadingHandler);
	viewer.addEventHandler(new osgViewer::HelpHandler);


	int _video_id = osgART::PluginManager::instance()->load("osgart_video_artoolkit2");
	int _tracker_id = osgART::PluginManager::instance()->load("osgart_tracker_artoolkit2");

	// Load a video plugin.
	video = dynamic_cast<osgART::Video*>(osgART::PluginManager::instance()->get(_video_id));
	if (!video.valid()){   
		notify(FATAL) << "Could not initialize video plugin!" << std::endl;
		exit(-1);
	}

	// found video - configure now
	osgART::VideoConfiguration* _config = video->getVideoConfiguration();

	// if the configuration is existing
	/*if (_config){
		_config->deviceconfig = "<?xml version=\"1.0\" encoding=\"UTF-8\"?>"
			"<dsvl_input><avi_file use_reference_clock=\"true\" file_name=\"hmd2.avi\" loop_avi=\"true\" render_secondary=\"false\">"
			"<pixel_format><RGB32/></pixel_format></avi_file></dsvl_input>";
	}*/
	//video->setFlip(false,true);

	video->open();// Open the video. Get information about the format of the video from user

	ref_ptr<osgART::Tracker> tracker = dynamic_cast<osgART::Tracker*>(osgART::PluginManager::instance()->get(_tracker_id));
	if (!tracker.valid()){
		notify(FATAL) << "Could not initialize tracker plugin!" << std::endl;
		exit(-1);
	}

	ref_ptr<osgART::Calibration> calibration = tracker->getOrCreateCalibration();
	if (!calibration->load(std::string("camera_para.dat"))){
		notify(FATAL) << "Non existing or incompatible calibration file" << std::endl;
		exit(-1);
	}

	
	tracker->setImage(video.get());// set the image source for the tracker

	osgART::TrackerCallback::addOrSet(root.get(), tracker.get());

	marker = tracker->addMarker("single;patt.kanji;80;0;0");
	if (!marker.valid()){
		// Without marker an AR application can not work. Quit if none found.
		notify(FATAL) << "Could not add marker!" << std::endl;
		exit(-1);
	}


	marker->setActive(true);

	MatrixTransform *arTransform = new MatrixTransform();

	osgART::attachDefaultEventCallbacks(arTransform,marker.get());

	arTransform->getOrCreateStateSet()->setRenderBinDetails(100, "RenderBin");

	ref_ptr<Group> videoBackground = createImageBackground(video.get());
	videoBackground->getOrCreateStateSet()->setRenderBinDetails(0, "RenderBin");

	ref_ptr<Camera> cam = calibration->createCamera();

	float scale =5;
	float height =103;
	transformMatrix->setMatrix(osg::Matrix::identity());
	transformMatrix->setMatrix(osg::Matrix::translate(Vec3d(0,height,0))*
		osg::Matrix::scale(scale, scale,1)*
		osg::Matrix::rotate(-18*_PI/180.0, 1.0f, 0.0f, 0.0f));

	transformMatrix->setDataVariance(osg::Object::DYNAMIC);
	transformMatrix->addChild(trainerAnimation.getGeode());
	arTransform->addChild(transformMatrix.get());

	cam->addChild(arTransform);
	cam->addChild(videoBackground.get());

	root->addChild(cam.get());

	video->start();
	viewer.realize();
	viewer.frame();
}

int main(void){
	trainerAnimation.init();
	setUpGraphics();

	Athlete athlete("Player One");
	Trainer trainer("John");

	athlete.init();

	athlete.addObserver(trainer);
	trainer.pose.addObserver(trainerAnimation);

	trainer.pose.notifyObservers();

	DataVisitor *dataVisitor = new XMLVisitor();
	athlete.acceptExporter(*dataVisitor);
	trainer.acceptExporter(*dataVisitor);

	delete dataVisitor;

	double feedbackTime=clock();
	while(!athlete.pressedQuit() || !viewer.done()){
		Sleep(5);

		athlete.updateInfoFromMarker(osg::Matrix::identity() != marker->getTransform() && marker->valid(), marker->getTransform().getTrans().y());	
		
		hud->setText(trainer.getFeedback());
		if((double)clock() > feedbackTime + (2 * CLOCKS_PER_SEC)){
			feedbackTime=(double)clock();

			cout << trainer.getFeedback() << endl;
			trainer.clearFeedback();
		}

		const TrainerPose *trainerPose = dynamic_cast<const TrainerPose *>(trainer.getPose());
		if(trainerPose->rightPunching()||trainerPose->leftPunching()){
			athlete.hit_=true;
		}else{
			athlete.hit_=false;
		}

		viewer.frame();
	}

	video->close();
	return 0;
}
